use crate::config;
use log;
use matrix_sdk::{
    config::SyncSettings,
    ruma::{
        events::room::{
            member::StrippedRoomMemberEvent,
            message::{RoomMessageEventContent, SyncRoomMessageEvent},
        },
        OwnedRoomId, TransactionId, UserId,
    },
    Client, Room,
};
use tokio::sync::mpsc::Receiver;
use tokio::time::{sleep, Duration};

pub(crate) async fn build(
    config: config::MatrixConfig,
) -> Result<matrix_sdk::Client, anyhow::Error> {
    let user = UserId::parse(config.username)
        .expect("invalid matrix username - should be a full mxid with server");
    let client = Client::builder()
        .server_name(user.server_name())
        .build()
        .await?;

    client
        .matrix_auth()
        .login_username(user, &config.password)
        .initial_device_display_name(&config.device_name)
        .send()
        .await?;

    client.add_event_handler(|ev: SyncRoomMessageEvent| async move {
        log::debug!("[matrix] Received a message {:?}", ev);
    });

    client.add_event_handler(on_stripped_state_member);

    log::debug!("connected to matrix");

    return Ok(client);
}

async fn on_stripped_state_member(
    room_member: StrippedRoomMemberEvent,
    client: Client,
    room: Room,
) {
    if room_member.state_key != client.user_id().unwrap() {
        return;
    }

    tokio::spawn(async move {
        log::debug!("Joining room {}", room.room_id());
        let mut delay = 2;

        while let Err(err) = room.join().await {
            // retry autojoin due to synapse sending invites, before the
            // invited user can join for more information see
            // https://github.com/matrix-org/synapse/issues/4345
            log::error!(
                "Failed to join room {} ({err:?}), retrying in {delay}s",
                room.room_id(),
            );

            sleep(Duration::from_secs(delay)).await;
            delay *= 2;

            if delay > 3600 {
                log::error!("Can't join room {} ({err:?})", room.room_id());
                break;
            }
        }
        log::debug!("Successfully joined room {}", room.room_id());

        let content = RoomMessageEventContent::text_plain("fuck you");
        let txn_id = TransactionId::new();

        let result = room.send(content).with_transaction_id(&txn_id).await;
        match result {
            Ok(_) => log::error!("sent hello message"),
            Err(err) => log::error!(
                "Error sending hello message to room {}: {err:?}",
                room.room_id()
            ),
        }
    });
}

pub(crate) async fn sender(
    client: Client,
    mut rx: Receiver<RoomMessageEventContent>,
    room_id: OwnedRoomId,
) {
    let room = match client.get_room(&room_id) {
        Some(room) => room,
        None => panic!("requested matrix room not found"),
    };

    while let Some(msg) = rx.recv().await {
        match room.send(msg).await {
            Err(err) => log::error!("error sending message to matrix: {:?}", err),
            Ok(response) => log::debug!("sent message to matrix: {:?}", response),
        }
    }
}

pub(crate) async fn sync(client: Client) {
    client
        .sync(SyncSettings::default())
        .await
        .expect("error syncing matrix");
}
